/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package main.tools;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author xmic
 */
public class KNNResultFileParser {

    private final List<EvaluatedQuery> evaluatedQueries;
    private final Integer candidateSetSize;

    public KNNResultFileParser(String filePath) {
        this(filePath, -1);
    }

    public KNNResultFileParser(String filePath, int k) {
        String name = new File(filePath).getName();
        if (name.startsWith("approx-k")) {
            candidateSetSize = Integer.parseInt(name.substring(name.indexOf("accessed") + "accessed".length(), name.lastIndexOf(".")));
        } else if (!name.contains("ground")){
            candidateSetSize = selectDigitsFromString(name);
        }else{
            candidateSetSize = null;
        }
        evaluatedQueries = new ArrayList<>();
        parse(filePath, k);
    }

    private void parse(String filePath, int k) {
        if (k < 0) {
            k = Integer.MAX_VALUE;
        }
        try {
            BufferedReader br = new BufferedReader(new FileReader(filePath));
            String locatorURI = null;
            List<String> nearestNeighbours = null;
            List<Float> distances = null;
            String firstLine = null;
            Integer time = null;
            Integer distanceComputations = null;
            try {
                while (true) {
                    String line = br.readLine();
                    if (line.startsWith("OperationTime")) {
                        String timeS = line.substring(line.indexOf(" ") + 1);
                        time = Integer.parseInt(timeS);
                    }
                    if (line == null || line.contains("returned") || line.contains("query=")) {
                        if (locatorURI != null) {
                            EvaluatedQuery evaluatedQuery = new EvaluatedQuery(distanceComputations, firstLine, locatorURI, nearestNeighbours, distances, candidateSetSize, time);
                            if (!evaluatedQueries.contains(evaluatedQuery)) {
                                evaluatedQueries.add(evaluatedQuery);
                            } else {
                                Logger.getLogger(KNNResultFileParser.class.getName()).log(Level.WARNING, "Query evaluated again: {0}", new Object[]{locatorURI});
                            }
                        }
                        firstLine = line;
                        if (line.contains("returned")) {
                            locatorURI = line.substring(line.indexOf("(") + 1, line.indexOf(")"));
                        } else {
                            locatorURI = line.substring(line.indexOf("query=") + "query=".length());
                        }
                        nearestNeighbours = new ArrayList<>();
                        distances = new ArrayList<>();
                        time = null;
                    }
                    if (Character.isDigit(line.charAt(0))) {
                        String[] split = line.split(", ");
//                        System.err.println(locatorURI);
                        for (int i = 0; i < Math.min(k, split.length); i++) {
//                            System.err.println(split[i]);
                            String[] pair = split[i].split(": ");
                            nearestNeighbours.add(pair[1]);
                            distances.add(Float.parseFloat(pair[0]));
                        }
                    }
                    if (line.contains("distance computations")) {
                        String distCompS = line.substring(line.indexOf("distance computations") + "distance computations".length() + 2);
                        distanceComputations = Integer.parseInt(distCompS);
                    }
                }
            } catch (NullPointerException ex) {
                evaluatedQueries.add(new EvaluatedQuery(distanceComputations, firstLine, locatorURI, nearestNeighbours, distances, candidateSetSize, time));
                if (!evaluatedQueries.isEmpty() || candidateSetSize != null) {
                    Logger.getLogger(KNNResultFileParser.class.getName()).log(Level.INFO, "Candidate set size: {0}, found {1} evaluated queries, first with {2} neighbours.", new Object[]{candidateSetSize, evaluatedQueries.size(), evaluatedQueries.get(0).nearestNeighbours.size()});
                }
            }
        } catch (IOException ex) {
            Logger.getLogger(KNNResultFileParser.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public List<EvaluatedQuery> getEvaluatedQueries() {
        return evaluatedQueries;
    }

    public int getNumberOfEvaluatedQueries() {
        return evaluatedQueries.size();
    }

    public Integer getCandidateSetSize() {
        return candidateSetSize;
    }

    private Integer selectDigitsFromString(String s) {
        String ret = "";
        for (int i = 0; i < s.length(); i++) {
            char charAt = s.charAt(i);
            if (Character.isDigit(charAt)) {
                ret += charAt;
            }
        }
        if ("".equals(ret)) {
            return null;
        }
        return Integer.parseInt(ret);
    }

    public static class EvaluatedQuery {

        private final String locatorURI;
        private final List<Integer> nearestNeighbours;
        private final List<Float> distances;
        private final Integer candidateSetSize;
        private final Integer time;
        private final Integer distanceComputations;
        private final String firstLine;

        public EvaluatedQuery(Integer distanceComputations, String firstLine, String locatorURI, List<String> nearestNeighbours, List<Float> distances, Integer candidateSetSize) {
            this(distanceComputations, firstLine, locatorURI, nearestNeighbours, distances, candidateSetSize, null);
        }

        public EvaluatedQuery(Integer distanceComputations, String firstLine, String locatorURI, List<String> nearestNeighbours, List<Float> distances, Integer candidateSetSize, Integer time) {
            this.locatorURI = locatorURI;
            this.nearestNeighbours = Tools.transformStringsToIntegers(nearestNeighbours);
            this.distances = distances;
            this.candidateSetSize = candidateSetSize;
            this.distanceComputations = distanceComputations;
            this.time = time;
            this.firstLine = firstLine;
        }
        
        

        public String getLocatorURI() {
            return locatorURI;
        }

        public List<Integer> getNearestNeighbours() {
            return nearestNeighbours;
        }

        public Set<Integer> getNearestNeighboursSet() {
            return new HashSet<>(nearestNeighbours);
        }

        public List<Float> getDistances() {
            return distances;
        }

        public Integer getCandidateSetSize() {
            return candidateSetSize;
        }

        public String getFirstLine() {
            return firstLine;
        }

        public Integer getTime() {
            return time;
        }

        public Integer getDistanceComputations() {
            return distanceComputations;
        }

        @Override
        public int hashCode() {
            try {
                int parseInt = Integer.parseInt(locatorURI);
                return Integer.hashCode(parseInt);
            } catch (NumberFormatException ex) {
                return locatorURI.hashCode();
            }
        }

        @Override
        public boolean equals(Object obj
        ) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final EvaluatedQuery other = (EvaluatedQuery) obj;
            try {
                Integer thisInt = Integer.parseInt(locatorURI);
                Integer oInt = Integer.parseInt(other.locatorURI);
                return thisInt.equals(oInt);
            } catch (NumberFormatException ex) {
                return locatorURI.equals(other.locatorURI);
            }
        }

        public void print(String prefix, PrintStream out) {
            if (prefix == null) {
                out.println(firstLine);
            } else {
                out.println(prefix + locatorURI);
            }
            for (int i = 0; i < nearestNeighbours.size(); i++) {
                if (i > 0) {
                    out.print(", ");
                }
                Float dist = distances.get(i);
                Integer locator = nearestNeighbours.get(i);
                out.print(dist + ": " + locator);
            }
            out.println();
        }

    }

}
