package sf.algorithm;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;
import java.util.logging.Logger;
import messif.algorithms.Algorithm;
import messif.algorithms.AlgorithmMethodException;
import messif.algorithms.impl.ParallelSequentialScan;
import messif.buckets.impl.MemoryStorageBucket;
import messif.objects.LocalAbstractObject;
import messif.objects.impl.MessifMapDBSerializer;
import messif.operations.GetCandidateSetOperation;
import messif.operations.data.BulkInsertOperation;
import messif.operations.query.ApproxKNNQueryOperation;
import messif.operations.query.GetAlgorithmInfoOperation;
import messif.operations.query.GetAllObjectsQueryOperation;
import messif.operations.query.KNNQueryOperation;
import messif.operations.query.RefineCandidateOperation;
import sf.objects.impl.Sketch;
import sf.objects.sketchcreator.ObjectToSketchTransformator;
import simcloud.idstorage.mapdb.AbstractMapDBRefiner;
import simcloud.idstorage.mapdb.FileMapDBRefiner;

/**
 *
 * @author xmic
 */
public class SketchAlgorithmWithRefinement extends Algorithm {

    public static final long serialVersionUID = -3557754663250874480l;
    protected final ParallelSequentialScan parSequentialScan;
    protected final ObjectToSketchTransformator objectToSketchTranformator;
    protected int candidatesK;
    protected final AbstractMapDBRefiner mapDB;
    /**
     * A simple lock for periodic algorithm serialization.
     */
    protected transient volatile Lock serializingLock = new ReentrantLock();
    protected transient volatile boolean modified = false;

    @Algorithm.AlgorithmConstructor(description = "Constructor from a created properties object", arguments = {"Algorithm name", "S-Index properties", "prefix"})
    public SketchAlgorithmWithRefinement(String name, Properties properties) throws Exception {
        super(name);
        String pivotFile = properties.getProperty("pivotFile");
        String objectClass = properties.getProperty("objectClass");
        int finalSketchSize = Integer.parseInt(properties.getProperty("finalSketchSize"));
        String csvFileWithPivotInfo = properties.getProperty("csvFileWithPivotInfo");
        boolean uniquePivots = Boolean.parseBoolean(properties.getProperty("uniquePivots"));
        float desiredCorrelation = Float.parseFloat(properties.getProperty("desiredCorrelation"));
        String sampleSetFile = properties.getProperty("sampleSetFile");
        Constructor<?> constructor = Class.forName(properties.getProperty("objectToSketchTranformatorClass")).getConstructor(String.class, String.class, int.class, String.class, boolean.class);
        if (sampleSetFile != null && properties.containsKey("bitCandidatesCountForCorrelation")) {
            int bitCandidateCount = Integer.parseInt(properties.getProperty("bitCandidatesCountForCorrelation"));
            objectToSketchTranformator = (ObjectToSketchTransformator) constructor.newInstance(pivotFile, objectClass, bitCandidateCount, csvFileWithPivotInfo, uniquePivots);
            objectToSketchTranformator.removeCorrelatedBits(finalSketchSize, sampleSetFile, objectClass, desiredCorrelation);
        } else {
            objectToSketchTranformator = (ObjectToSketchTransformator) constructor.newInstance(pivotFile, objectClass, finalSketchSize, csvFileWithPivotInfo, uniquePivots);
        }
        parSequentialScan = new ParallelSequentialScan(Integer.parseInt(properties.getProperty("paralelization")), MemoryStorageBucket.class);
        candidatesK = Integer.parseInt(properties.getProperty("candidatesK"));
        boolean useFullMapper = Boolean.parseBoolean(properties.getProperty("useFullMapper"));
        mapDB = new FileMapDBRefiner(properties.getProperty("mapDBStorageFile"), new MessifMapDBSerializer(), useFullMapper, false);
        modified = true;
    }

    @Algorithm.AlgorithmConstructor(description = "Constructor with specified arguments", arguments = {"objectToSketchTranformator", "paralelization", "candidatesK", "mapDBStorageFile"})
    public SketchAlgorithmWithRefinement(ObjectToSketchTransformator objectToSketchTranformator, int paralelization, int candidatesK, String mapDBStorageFile, boolean useFullMapper) {
        super("SketchAlgorithm");
        this.objectToSketchTranformator = objectToSketchTranformator;
        parSequentialScan = new ParallelSequentialScan(paralelization, MemoryStorageBucket.class);
        this.candidatesK = candidatesK;
        mapDB = new FileMapDBRefiner(mapDBStorageFile, new MessifMapDBSerializer(), useFullMapper, false);
        modified = true;
    }

    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
    }

    /**
     * Stores the whole algorithm into the passed file if the structure of the
     * dynamic cell tree was changed. This method is typically called
     * periodically (every 1s).
     *
     * @param serializationFile file to store the algorithm to
     * @throws java.io.IOException
     */
    public void checkModifiedAndStore(String serializationFile) throws IOException {
        serializingLock.lock();
        try {
            if (modified) {
                this.storeToFile(serializationFile);
                modified = false;
            }
        } finally {
            serializingLock.unlock();
        }
    }

    public void insertObjects(BulkInsertOperation op) {
        try {
            op = mapDB.executeOperation(op);;
            Iterator<? extends LocalAbstractObject> insertedObjects = op.getInsertedObjects().iterator();
            List<Sketch> sketches = new ArrayList<>();
            while (insertedObjects.hasNext()) {
                LocalAbstractObject objToAdd = insertedObjects.next();
                Sketch sketch = objectToSketchTranformator.transform(objToAdd);
                sketches.add(sketch);
            }
            BulkInsertOperation insertSketches = new BulkInsertOperation(sketches);
            parSequentialScan.executeOperation(insertSketches);
            insertSketches.endOperation();
            op.endOperation();
            modified = true;
        } catch (AlgorithmMethodException | NoSuchMethodException ex) {
            Logger.getLogger(SketchAlgorithmWithRefinement.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public KNNQueryOperation knnSearch(KNNQueryOperation op) {
        try {
            GetCandidateSetOperation candidateSetOperation = new GetCandidateSetOperation(op, candidatesK);
            candidateSetOperation = executeOperation(candidateSetOperation);
            RefineCandidateOperation refineOp = new RefineCandidateOperation(candidateSetOperation, op);
            refineOp = mapDB.executeOperation(refineOp);
            refineOp.endOperation();
            return op;
        } catch (AlgorithmMethodException | NoSuchMethodException ex) {
            Logger.getLogger(SketchAlgorithmWithRefinement.class.getName()).log(Level.SEVERE, null, ex);
        }
        return null;
    }

    public KNNQueryOperation knnApproxSearch(ApproxKNNQueryOperation op) {
        candidatesK = (int) op.getArgument(2);
//        try {
        //
//            Sketch querySketch = objectToSketchTranformator.transform(op.getQueryObject());
//            KNNQueryOperation knn = new KNNQueryOperation(querySketch, candidatesK);
//            knn = parSequentialScan.executeOperation(knn);
//            knn.endOperation();
//            op.endOperation();
//            return op;
        return knnSearch(op);
//        } catch (NoSuchMethodException ex) {
//            Logger.getLogger(SketchAlgorithmWithRefinement.class.getName()).log(Level.SEVERE, null, ex);
//        }
//        return null;
    }

    public GetCandidateSetOperation candidateSetOp(GetCandidateSetOperation op) {
        try {
            LocalAbstractObject queryObject = op.getEncapsulatedOperation().getQueryObject();
            Sketch querySketch = objectToSketchTranformator.transform(queryObject);
            KNNQueryOperation candPreselection = new KNNQueryOperation(querySketch, candidatesK);
            candPreselection = parSequentialScan.executeOperation(candPreselection);
            op.addAll(candPreselection.getAnswerObjects());
            candPreselection.endOperation();
            op.endOperation();
//            KNNApproxEvaluation.printAnswer(candPreselection, System.out);
            return op;
        } catch (AlgorithmMethodException | NoSuchMethodException ex) {
            Logger.getLogger(SketchAlgorithmWithRefinement.class.getName()).log(Level.SEVERE, null, ex);
        }
        return null;
    }

    @Override
    public void setOperationsThreadPool(ExecutorService operationsThreadPool) {
        super.setOperationsThreadPool(operationsThreadPool);
        mapDB.setOperationsThreadPool(operationsThreadPool);
    }

    public final void printAlgorithmInfo() {
        try {
            GetAlgorithmInfoOperation algorithmInfoOperation = new GetAlgorithmInfoOperation();
            parSequentialScan.executeOperation(algorithmInfoOperation);
            System.out.println(algorithmInfoOperation.getAnswer());
            System.out.println("objectToSketchTranformator class: " + objectToSketchTranformator.getClass().getName());
        } catch (AlgorithmMethodException | NoSuchMethodException ex) {
            Logger.getLogger(SketchAlgorithmWithRefinement.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    @Override
    @SuppressWarnings({"FinalizeDeclaration", "FinalizeCalledExplicitly"})
    public void finalize() throws Throwable {
        if (mapDB != null) {
            mapDB.finalize();
        }
        super.finalize();
    }

    @Override
    public void destroy() throws Throwable {
        if (mapDB != null) {
            mapDB.destroy();
        }
        super.destroy();
    }

    public int getCandidatesK() {
        return candidatesK;
    }

    public void setCandidatesK(int candidatesK) {
        this.candidatesK = candidatesK;
    }

    public GetAllObjectsQueryOperation getAllSketches(GetAllObjectsQueryOperation op) {
        try {
            return parSequentialScan.executeOperation(op);
        } catch (AlgorithmMethodException | NoSuchMethodException ex) {
            Logger.getLogger(SketchAlgorithmWithRefinement.class.getName()).log(Level.SEVERE, null, ex);
        }
        return null;
    }

    public ParallelSequentialScan getParSequentialScan() {
        return parSequentialScan;
    }

}
