package sf.pivotAnalyzer;

import java.io.BufferedReader;
import java.io.EOFException;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import messif.objects.LocalAbstractObject;

/**
 *
 * @author xmic
 */
public abstract class CsvPivotInfoFileProcessor {

    private final String csvFilePath;
    private final Map<String, LocalAbstractObject> pivots;
    private final LocalAbstractObject[] pivotsArray;
    private final String delimiter;
    private final boolean uniquePivots;

    public CsvPivotInfoFileProcessor(String csvFilePath, boolean uniquePivots, LocalAbstractObject[] pivots, String delimiter) {
        this.csvFilePath = csvFilePath;
        this.delimiter = delimiter;
        this.pivotsArray = pivots;
        this.uniquePivots = uniquePivots;
        this.pivots = new HashMap<>();
        for (int i = 0; i < pivots.length; i++) {
            this.pivots.put(pivots[i].getLocatorURI(), pivots[i]);
        }
    }

    public CsvPivotInfoFileProcessor(String csvFilePath, boolean uniquePivots, LocalAbstractObject[] pivots) {
        this(csvFilePath, uniquePivots, pivots, ";");
    }

    public CsvPivotInfoFileProcessor(String csvFilePath, LocalAbstractObject[] pivots) {
        this(csvFilePath, false, pivots);
    }

    public LocalAbstractObject[] getPivotsAccordingCsvLocators(int numberOfPivots) {
        if (csvFilePath == null || !new File(csvFilePath).exists()) {
            Logger.getLogger(CsvPivotInfoFileProcessor.class.getName()).log(Level.WARNING, "Csv file with pivot locators pairs is not provided or doesn''t exists. Pivots in a native order is used. File path:\n{0}", csvFilePath);
            LocalAbstractObject[] ret = new LocalAbstractObject[numberOfPivots];
            System.arraycopy(pivotsArray, 0, ret, 0, numberOfPivots);
            return ret;
        }
        BufferedReader br = null;
        if (numberOfPivots < 0) {
            numberOfPivots = pivotsArray.length * (pivotsArray.length - 1) / 2;
        }
        try {
            br = new BufferedReader(new FileReader(csvFilePath));
            List<LocalAbstractObject> ret = new ArrayList();
            int rowCounter = 0;
            try {
                while (ret.size() < numberOfPivots) {
                    String line = br.readLine();
                    if (line == null) {
                        Logger.getLogger(CsvPivotInfoFileProcessor.class.getName()).log(Level.WARNING, "Found only {0} pivot pairs.", ret.size());
                        break;
                    }
                    String[] split = line.split(delimiter);
                    String[] locators = processSplitLine(split, rowCounter);
                    rowCounter++;
                    boolean add = true;
                    List<LocalAbstractObject> added = new ArrayList<>();
                    for (int i = 0; i < locators.length; i++) {
                        LocalAbstractObject cand = getPivotByLocator(removeQutes(locators[i]));
                        if (uniquePivots && ret.contains(cand)) {
                            add = false;
                            break;
                        }
                        added.add(cand);
                    }
                    if (add) {
                        ret.addAll(added);
                    }
                }
            } catch (EOFException ex) {
                Logger.getLogger(CsvPivotInfoFileProcessor.class.getName()).log(Level.WARNING, null, ex);
            }
            Logger.getLogger(CsvPivotInfoFileProcessor.class.getName()).log(Level.INFO, "Csv file {0} used to make pivot pairs. Returned {1} pivots.", new Object[]{csvFilePath, ret.size()});
            return ret.toArray(new LocalAbstractObject[0]);
        } catch (IOException ex) {
            Logger.getLogger(CsvPivotInfoFileProcessor.class.getName()).log(Level.SEVERE, null, ex);
        } finally {
            try {
                br.close();
            } catch (IOException ex) {
                Logger.getLogger(CsvPivotInfoFileProcessor.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        return null;
    }

    protected String removeQutes(String string) {
        if (string.startsWith("\"") && string.endsWith("\"")) {
            return string.substring(1, string.length() - 1);
        }
        return string;
    }

    protected LocalAbstractObject getPivotByLocator(String locator) {
        if (!pivots.containsKey(locator)) {
            Logger.getLogger(CsvPivotInfoFileProcessor.class.getName()).log(Level.SEVERE, "Map of pivots doesn''t contain locator {0}. Map size: {1}. First: {2}", new Object[]{locator, pivots.size(), pivots.keySet().iterator().next()});
        }
        return pivots.get(locator);
    }

    /**
     * Return pivot locator (one or more) from the line. Other pieces of
     * information must be hold in an inhereted class.
     *
     * @param split cells of csv line
     * @param index row index
     * @return pivot locator (one or more) from the line.
     */
    protected abstract String[] processSplitLine(String[] split, int index);
}
