package sf.pivotAnalyzer;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import main.tools.Tools;
import messif.objects.LocalAbstractObject;
import messif.objects.util.AbstractObjectList;
import messif.objects.util.StreamGenericAbstractObjectIterator;
import sf.pivotAnalyzer.threads.PivotCoupleAnalyzer;
import sf.pivotAnalyzer.threads.PivotCoupleAnalyzer.IntArraySameLengthComparator;
import sf.pivotAnalyzer.threads.impl.HyperplaneDistancesAnalyzerGivenPairs;
import sf.pivotAnalyzer.threads.impl.HyperplanePartsSizeRatioAnalyzerAllCouples;

/**
 *
 * @author xmic
 */
public class PivotSubsetSelector {

    public static final Integer PARALELIZATION = 8;

    private final LocalAbstractObject[] pivots;
    private final Iterator<LocalAbstractObject> pivotsIt;

    public PivotSubsetSelector(LocalAbstractObject[] pivots) {
        this.pivots = pivots;
        pivotsIt = null;
    }

    public PivotSubsetSelector(Iterator<LocalAbstractObject> pivotsIt) {
        this.pivotsIt = pivotsIt;
        pivots = (LocalAbstractObject[]) new AbstractObjectList(pivotsIt).toArray(new LocalAbstractObject[0]);
    }

    public LocalAbstractObject[] getFirstPivots(int numberOfReturnedPivots) {
        if (pivots != null) {
            return Arrays.copyOf(pivots, numberOfReturnedPivots);
        }
        return (LocalAbstractObject[]) new AbstractObjectList(pivotsIt, numberOfReturnedPivots).toArray(new LocalAbstractObject[0]);
    }

    /**
     * Returns map with key of length 2, containing pivots indexes, and value
     * equal to the ratio of sample set objects, closer to the first pivot. All
     * possible couples are investigated. It means i from [0, pivots.length - 1)
     * and j from [i + 1,pivots.length). Values equal to zero are not contained.
     *
     * @param sampleSetIterator
     * @param minorPartSizeRatioUpperBound
     * @param minorPartSizeRatioLowerBound threshold for the minor part size.
     * @return
     */
    public Map<LocalAbstractObject[], Float> getPivotsWithProperHyperplaneDivision(float minorPartSizeRatioLowerBound, float minorPartSizeRatioUpperBound, Iterator<LocalAbstractObject> sampleSetIterator) {
        Map<LocalAbstractObject[], Float> ret = new HashMap<>();
        Map<int[], Float> dividingRatiosMap = countCouplesStatsMap(pivots, sampleSetIterator, minorPartSizeRatioLowerBound, minorPartSizeRatioUpperBound, HyperplanePartsSizeRatioAnalyzerAllCouples.class);
        for (Map.Entry<int[], Float> entry : dividingRatiosMap.entrySet()) {
            int[] key = entry.getKey();
            ret.put(new LocalAbstractObject[]{pivots[key[0]], pivots[key[1]]}, entry.getValue());
        }
        Logger.getLogger(PivotSubsetSelector.class.getName()).log(Level.INFO, "Pivot selector returned {0} pivot couples with hyperplane division better than {1}/{2}.", new Object[]{ret.size(), minorPartSizeRatioLowerBound, 1 - minorPartSizeRatioLowerBound});
        return ret;
    }

    public void getPivotsDistancesFromHyperplane(StreamGenericAbstractObjectIterator<LocalAbstractObject> sampleSetIterator, Iterator<LocalAbstractObject> pivotsIterator) {
        try {
            while (pivotsIterator.hasNext()) {
                List<LocalAbstractObject> pivotsBatch = new AbstractObjectList<>(pivotsIterator, 30000);
                sampleSetIterator.reset();
                Map<LocalAbstractObject[], Float> map = getPivotsDistancesFromHyperplaneBatch(sampleSetIterator, pivotsBatch.toArray(new LocalAbstractObject[0]));
                Tools.print(map, (OutputStream) null);
            }
        } catch (IOException ex) {
            Logger.getLogger(PivotSubsetSelector.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    private Map<LocalAbstractObject[], Float> getPivotsDistancesFromHyperplaneBatch(Iterator<LocalAbstractObject> sampleSetIterator, LocalAbstractObject[] pivotsBatch) {
        Map<LocalAbstractObject[], Float> ret = new HashMap<>();
        Map<int[], Float> map = countCouplesStatsMap(pivotsBatch, sampleSetIterator, -1, -1, HyperplaneDistancesAnalyzerGivenPairs.class);
        for (Map.Entry<int[], Float> entry : map.entrySet()) {
            int[] key = entry.getKey();
            ret.put(new LocalAbstractObject[]{pivotsBatch[key[0]], pivotsBatch[key[1]]}, entry.getValue());
        }
        Logger.getLogger(PivotSubsetSelector.class.getName()).log(Level.INFO, "Pivot selector returned {0} pivot couples.", new Object[]{ret.size()});
        return ret;
    }

    private Map<int[], Float> countCouplesStatsMap(LocalAbstractObject[] pivots, Iterator<LocalAbstractObject> sampleSetIterator, float minorPartRatioLowerThreshold, float minorPartRatioUpperThreshold, Class pivotCoupleAnalyzerClass) {
        try {
            Constructor<PivotCoupleAnalyzer> constructor = pivotCoupleAnalyzerClass.getConstructor(LocalAbstractObject[].class, List.class);
            int size = 0;
            SortedMap<int[], Float> candidatesAll = new TreeMap<>(new IntArraySameLengthComparator());
            while (sampleSetIterator.hasNext()) {
                PivotCoupleAnalyzer[] threads = new PivotCoupleAnalyzer[PARALELIZATION];
                for (int i = 0; i < PARALELIZATION; i++) {
                    List<LocalAbstractObject> sampleBatch = new AbstractObjectList<>(sampleSetIterator, 1000);
                    if (sampleBatch.isEmpty()) {
                        threads[i] = null;
                    } else {
                        size += sampleBatch.size();
                        PivotCoupleAnalyzer thread = constructor.newInstance(pivots, sampleBatch);
                        threads[i] = thread;
                        Logger.getLogger(PivotSubsetSelector.class.getName()).log(Level.INFO, "Pivot selector map thread {0} start. Sample batch size: {1}, pivots size: {2}", new Object[]{thread.getId(), sampleBatch.size(), pivots.length});
                        thread.start();
                    }
                }
                for (Thread th : threads) {
                    if (th != null) {
                        th.join();
                    }
                }
                System.gc();
                for (PivotCoupleAnalyzer th : threads) {
                    if (th == null) {
                        continue;
                    }
                    Map<int[], Float> result = th.getResult();
                    for (Map.Entry<int[], Float> entry : result.entrySet()) {
                        int[] key = entry.getKey();
                        Float value = entry.getValue();
                        if (candidatesAll.containsKey(key)) {
                            candidatesAll.put(key, candidatesAll.get(key) + value);
                        } else {
                            candidatesAll.put(key, value);
                        }
                    }
                }
            }
            boolean addAll = minorPartRatioLowerThreshold < 0;
            if (minorPartRatioUpperThreshold <= minorPartRatioLowerThreshold) {
                minorPartRatioUpperThreshold = 0.5f;
            }
            SortedMap<int[], Float> ret = new TreeMap<>(new IntArraySameLengthComparator());
            for (Map.Entry<int[], Float> entry : candidatesAll.entrySet()) {
                int[] key = entry.getKey();
                float value = entry.getValue() / size;
                if (value < 1 - value) {
                    int s = key[0];
                    key[0] = key[1];
                    key[1] = s;
                }
                value = Math.min(value, 1 - value);
                if (addAll || (value >= minorPartRatioLowerThreshold && value <= minorPartRatioUpperThreshold)) {
                    ret.put(key, value);
                }
            }
            return ret;
        } catch (InterruptedException | NoSuchMethodException | SecurityException | InstantiationException | IllegalAccessException | IllegalArgumentException | InvocationTargetException ex) {
            Logger.getLogger(PivotSubsetSelector.class.getName()).log(Level.SEVERE, null, ex);
        }
        return null;
    }


    public static void printIndexes(int[] indexes, OutputStream out) {
        try {
            for (int i = 0; i < indexes.length; i++) {
                String s = "\"" + indexes[i] + "\"\n";
                out.write(s.getBytes());
            }
            out.flush();
        } catch (IOException ex) {
            Logger.getLogger(PivotSubsetSelector.class.getName()).log(Level.SEVERE, null, ex);
        }

    }

    public static void printPivots(LocalAbstractObject[] pivots, OutputStream out) {
        try {
            for (int i = 0; i < pivots.length; i++) {
                String s = pivots[i].getLocatorURI() + "\n";
                out.write(s.getBytes());
            }
        } catch (IOException ex) {
            Logger.getLogger(PivotSubsetSelector.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

}
