package sf.pivotAnalyzer.uncorrelatedSelector;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import sf.objects.impl.Sketch;

/**
 * Sketches are created as an object bit string represetnation. Bits should be
 * uncorrelated. When inverted sketches are build (bit relates to sample set
 * objects, sketch to given pivot(s)), this class chooses the sketch indexes
 * (i.e. pivot(s) indexes), which produce mutually uncorrelated sketches.
 *
 * @author xmic
 */
public abstract class InvertedSketchesCorrelationCounter {

    private final List<Sketch> invertedSketches;
    private final int sampleObjectNumber;
    protected final float correlationToAchieve;

    public InvertedSketchesCorrelationCounter(List<Sketch> invertedSketches, int sampleObjectNumber, float correlationToAchieve) {
        this.invertedSketches = invertedSketches;
        this.sampleObjectNumber = sampleObjectNumber;
        this.correlationToAchieve = correlationToAchieve;
    }

    /**
     * Returns matrix where field i,j contains absolute value of Pearson
     * correlation of sketches in positions i and j.
     *
     * @return
     */
    private float[][] getSketchesCorrelations(String filePath, int size) {
        float[][] ret = null;
        BufferedReader br = null;
        try {
            br = new BufferedReader(new FileReader(filePath));
            int counter = 0;
            if (size <= 0) {
                size = Integer.MAX_VALUE;
            }
            try {
                while (ret == null || counter < size) {
                    String[] floats = br.readLine().split(";");
                    if (ret == null) {
                        if (size == Integer.MAX_VALUE) {
                            size = floats.length;
                        }
                        ret = new float[size][size];
                    }
                    for (int i = 0; i < size; i++) {
                        float value = Float.parseFloat(floats[i]);
                        ret[counter][i] = value;
                        ret[i][counter] = value;
                    }
                    counter++;
                }
            } catch (NullPointerException ignore) {
            }
        } catch (IOException ex) {
            Logger.getLogger(InvertedSketchesCorrelationCounter.class.getName()).log(Level.SEVERE, null, ex);
        } finally {
            try {
                br.close();
            } catch (IOException ex) {
                Logger.getLogger(InvertedSketchesCorrelationCounter.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        return ret;
    }

    public float[][] getSketchesCorrelations(boolean printMap, boolean absoluteValues) {
        return this.getSketchesCorrelations(printMap, absoluteValues, null);
    }

    public float[][] getSketchesCorrelations(boolean printMap, boolean absoluteValues, Integer splitIntoParts) {
        List<float[]> sketchesCardSizeAvg = new ArrayList<>();
        float[][] ret = new float[invertedSketches.size()][invertedSketches.size()];
        for (int i = 0; i < invertedSketches.size(); i++) {
            sketchesCardSizeAvg.add(invertedSketches.get(i).getCardSizeCardCard2(sampleObjectNumber));
            ret[i][i] = 1f;
        }

        int blockSize = splitIntoParts == null ? invertedSketches.size() : invertedSketches.size() / splitIntoParts;

        long timeStart = System.currentTimeMillis();
        int start = 0;
        int end = blockSize - 1;
        while (end < invertedSketches.size()) {
            for (int i = start; i < end; i++) {
                for (int j = i + 1; j < end + 1; j++) {
                    float cor = invertedSketches.get(i).getPearsonCorrelation(invertedSketches.get(j), sketchesCardSizeAvg.get(i), sketchesCardSizeAvg.get(j), sampleObjectNumber);
                    if (absoluteValues) {
                        cor = Math.abs(cor);
                    }
                    ret[i][j] = cor;
                    ret[j][i] = cor;
//                    System.out.println(cor);;
                }
            }
            start += blockSize;
            end += blockSize;
        }
        Logger.getLogger(InvertedSketchesCorrelationCounter.class.getName()).log(Level.INFO, "{0} correlations counted in {1} ms.", new Object[]{invertedSketches.size() * (invertedSketches.size() - 1) / 2, System.currentTimeMillis() - timeStart});
        if (printMap) {
            for (int i = 0; i < invertedSketches.size() - 1; i++) {
                for (int j = 0; j < invertedSketches.size(); j++) {
//                for (int j = i + 1; j < invertedSketches.size(); j++) {
                    System.out.println(Float.toString(ret[i][j]) + ";");
                }
                System.out.println();
            }
        }
        return ret;
    }

    public int[] getIndexesOfLowCorrelatedPivots(int tuplesLength, int returnedBitsCount) {
        if (returnedBitsCount > invertedSketches.size()) {
            throw new IllegalArgumentException("Number of provided sketches is " + invertedSketches.size() + ". Tried to get " + returnedBitsCount + " sketches.");
        }
        if (returnedBitsCount <= 1 || returnedBitsCount >= invertedSketches.size()) {
            throw new IllegalArgumentException("This number of returned indexes is useless. Sketches size: " + invertedSketches.size() + ". Tried to get " + returnedBitsCount + " sketches.");
        }
//        float[][] sketchesCorrelations = getSketchesCorrelations("c:\\Data\\2015_Doktorske\\2015_Sketches\\Testy\\Korelace\\Tmp_6_invSk.csv", -1);;
//        float[][] sketchesCorrelations = getSketchesCorrelations("c:\\Data\\2015_Doktorske\\2015_Sketches\\Testy\\Korelace\\10000obj,1024pivots.csv", -1);;
        float[][] sketchesCorrelations = getSketchesCorrelations(false, false);
        return selectLowCorrelatedSubset(tuplesLength, sketchesCorrelations, returnedBitsCount);
    }

    protected abstract int[] selectLowCorrelatedSubset(int tuplesLength, float[][] sketchesCorrelations, int sketchSize);

    protected float getHighestCorrelation(Integer[] curIndexes, float[][] sketchesCorrelations) {
        float ret = 0;
        for (int i = 0; i < curIndexes.length - 1; i++) {
            int idxI = curIndexes[i];
            for (int j = i + 1; j < curIndexes.length; j++) {
                int idxJ = curIndexes[j];
                if (sketchesCorrelations[idxI][idxJ] > ret) {
                    ret = sketchesCorrelations[idxI][idxJ];
                }
            }
        }
        return ret;
    }

    protected void updateCorrelations(float[][] sketchesCorrelations, float correlationToAdd) {
        for (int i = 0; i < sketchesCorrelations.length; i++) {
            float[] sketchesCorrelationLine = sketchesCorrelations[i];
            for (int j = 0; j < sketchesCorrelationLine.length; j++) {
                float t = sketchesCorrelationLine[j];
                sketchesCorrelations[i][j] = t + correlationToAdd;
            }
        }
    }

}
