/**
 * *****************************************************************************
 * Copyright (c) 2013, Lucia Vadicamo (NeMIS Lab., ISTI-CNR, Italy) All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * ****************************************************************************
 */
package sf.separabilityAnalytique;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.math4.analysis.BivariateFunction;

/**
 * @author Lucia
 *
 */
public class BivariateIntegratorLucia {

    private final BivariateFunction func;
    private final double min;
    private final double max;
    private final double t;

    public final double step;

//    public BivariateIntegratorLucia(BivariateFunction func, double a, double b) {
//        this(func, a, b, b);
//    }
//
//    public BivariateIntegratorLucia(BivariateFunction func, double a, double b, double t) {
//        this(func, a, b, t, IMPLICIT_INTERVAL);
//    }
    public BivariateIntegratorLucia(BivariateFunction func, double min, double max, double t, double step) {
        this.func = func;
        this.min = min;
        this.max = max;
        if (t <= 0) {
            this.t = max;
        } else {
            this.t = t;
        }
        this.step = step;
    }

    public double evaluate() {
        checkParams();
        return evaluateWithStep(step);
    }

    private void checkParams() {
        boolean ok = max > min;
        ok = ok && func != null;
        if (!ok) {
            throw new IllegalArgumentException("Bad limits or function func: " + min + ", " + max + "; " + func);
        }
    }

    private double evaluateWithStep(double step) {
        // evaluator for case x1 from min to max and x2 from min to x1
        Logger.getLogger(BivariateIntegratorLucia.class.getName()).log(Level.INFO, "Trying to evaluate with step {0}", step);
        double ret = 0;
        double x2 = min;
        while (x2 < t) {
            double x1 = min;
            while (x1 < x2) {
                float numberOfCellsToParticipate;
                if (x1 == 0 && x2 == step) {
                    numberOfCellsToParticipate = 0;
                } else if (x2 == max - step && x1 == 0) {
                    numberOfCellsToParticipate = 1;
                } else {
                    float validNeighboursCount = getValidNeighboursCount(x1, x2, step);
                    if (validNeighboursCount == 1f) { // only [0,0]
                        numberOfCellsToParticipate = 1;
                    } else if (validNeighboursCount == 2f) {
                        numberOfCellsToParticipate = 2; // edges x1 = x2 
                    } else if (validNeighboursCount == 2.5f) {
                        numberOfCellsToParticipate = 2; // points next to corners [0,0] and [max, max]
                    } else if (validNeighboursCount == 3f) {
                        numberOfCellsToParticipate = 2;
                    } else if (validNeighboursCount == 3.5f) {
                        numberOfCellsToParticipate = 4f;
                    } else if (validNeighboursCount == 4f) {
                        numberOfCellsToParticipate = 4;
                    } else {
                        throw new RuntimeException("Not working properly. Bad cell with value " + validNeighboursCount);
                    }
                }
                float weight = numberOfCellsToParticipate / 4f;
                ret += weight * func.value(x1, x2);
                x1 += step;
            }
            x2 += step;
        }
        ret = ret * step * step;
        return ret;
    }

    private float getValidNeighboursCount(double x1, double x2, double step) {
        float ret = 0;
        ret += isValidPoint(x1 - step, x2, step);
        ret += isValidPoint(x1 + step, x2, step);
        ret += isValidPoint(x1, x2 - step, step);
        ret += isValidPoint(x1, x2 + step, step);
        return ret;
    }

    private float isValidPoint(double x1, double x2, double step) {
        if (x1 < 0 || x2 < 0) {
            return 0;
        }
        if (x1 >= max || x2 >= t) {
            return 0;
        }
        if (x1 < x2 - step) {
            return 1;
        }
        if (x1 == x2 - step) {
            return 0.75f;
        }
        return 0;
    }

}
