package sf.algorithm;

import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import messif.algorithms.Algorithm;
import messif.algorithms.AlgorithmMethodException;
import messif.buckets.BucketDispatcher;
import messif.buckets.BucketStorageException;
import messif.buckets.LocalBucket;
import messif.buckets.impl.DiskBlockBucket;
import messif.objects.LocalAbstractObject;
import messif.objects.util.AbstractObjectList;
import messif.operations.data.BulkInsertOperation;
import messif.operations.data.InsertOperation;
import messif.operations.query.KNNQueryOperation;
import sf.objects.impl.Sketch;
import sf.objects.sketchcreator.ObjectToSketchTransformator;
import sf.operations.data.GetObjectsBySketchesOperation;

/**
 *
 * @author xmic
 */
public class SketchObjectMapperAlgorithm extends Algorithm {

    public static final long serialVersionUID = -3557754663250874485l;

    private final Map<Sketch, LocalBucket> map;
    private final ObjectToSketchTransformator sketchCreator;
    private final String bucketFolderPath = "c:\\Data\\2015_Doktorske\\Desc\\buckets\\";
    private final BucketDispatcher bucketDispatcher;

    public SketchObjectMapperAlgorithm(ObjectToSketchTransformator sketchCreator) throws IllegalArgumentException {
        super("SketchObjectMapperAlgorithm");
        map = new HashMap<>();
        this.sketchCreator = sketchCreator;
        this.bucketDispatcher = new BucketDispatcher(Integer.MAX_VALUE, Long.MAX_VALUE, Long.MAX_VALUE, 0, false, DiskBlockBucket.class);
    }

    public void insert(InsertOperation insertOp) throws BucketStorageException {
        LocalAbstractObject insertedObject = insertOp.getInsertedObject();
        insertObject(insertedObject);
    }

    private void insertObject(LocalAbstractObject insertedObject) throws BucketStorageException {
        Sketch key = sketchCreator.transform(insertedObject);
        LocalBucket bucket;
        if (map.containsKey(key)) {
            bucket = map.get(key);
        } else {
            File file = new File(bucketFolderPath, Integer.toString(map.size()) + ".ds");
            Map<String, Object> params = new HashMap<>();
            params.put("file", file);
            bucket = bucketDispatcher.createBucket(DiskBlockBucket.class, params);
        }
        bucket.addObject(insertedObject);
        map.put(key, bucket);
    }

    public void insertObjects(BulkInsertOperation bulkInsertOp) throws BucketStorageException {
        List<? extends LocalAbstractObject> insertedObjects = bulkInsertOp.getInsertedObjects();
        for (LocalAbstractObject obj : insertedObjects) {
            insertObject(obj);
        }
        bulkInsertOp.endOperation();
    }

    public Collection<Sketch> getSketches() {
        return map.keySet();
    }

    public KNNQueryOperation knnSearch(KNNQueryOperation op) {
        try {
            AbstractObjectList candSketches = op.getParameter("candidates", AbstractObjectList.class);
            GetObjectsBySketchesOperation getCandidatesOp = new GetObjectsBySketchesOperation(candSketches);
            getCandidatesOp = executeOperation(getCandidatesOp);
            AbstractObjectList<LocalAbstractObject> candObjects = getCandidatesOp.getAnswer();
            op.evaluate(candObjects.iterator());
            op.endOperation();
            return op;
        } catch (AlgorithmMethodException | NoSuchMethodException ex) {
            Logger.getLogger(SketchObjectMapperAlgorithm.class.getName()).log(Level.SEVERE, null, ex);
        }
        return null;
    }

    public GetObjectsBySketchesOperation getObjectsBySketches(GetObjectsBySketchesOperation op) {
        AbstractObjectList<Sketch> sketches = op.getSketches();
        for (Sketch sketch : sketches) {
            if (map.containsKey(sketch)) {
                op.addToAnswer(map.get(sketch).getAllObjects());
            }
        }
        op.endOperation();
        return op;
    }

    @Override
    @SuppressWarnings({"FinalizeDeclaration", "FinalizeCalledExplicitly"})
    public void finalize() throws Throwable {
        if (bucketDispatcher != null) {
            bucketDispatcher.finalize();
        }
        super.finalize();
    }

    @Override
    public void destroy() throws Throwable {
        if (bucketDispatcher != null) {
            bucketDispatcher.destroy();
        }
    }

    private long getObjectBySketchTime = 0;
    private int candidatesSize = 0;

    public long resetGettingObjectBySketchTime() {
        long ret = getObjectBySketchTime;
        getObjectBySketchTime = 0;
        return ret;
    }

    public int resetGettingCandidatesSize() {
        int ret = candidatesSize;
        candidatesSize = 0;
        return ret;
    }

    public Iterator<LocalAbstractObject> getObjectsWithSketch(Sketch sketch) {
        getObjectBySketchTime -= System.currentTimeMillis();
        AbstractObjectList<LocalAbstractObject> ret = new AbstractObjectList<>();
        LocalBucket bucket = map.get(sketch);
        ret.addAll(bucket.getAllObjects());
        getObjectBySketchTime += System.currentTimeMillis();
        candidatesSize += ret.size();
        return ret.iterator();
    }
}
