package sf.objects.impl;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.BitSet;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import messif.objects.LocalAbstractObject;

/**
 *
 * @author xmic
 */
public class Sketch extends LocalAbstractObject implements Comparable<Sketch>, Serializable {

    public static final long serialVersionUID = 3402945598644229079l;

    /**
     * Java class for storing binary vectors
     */
    protected BitSet bitSet;

    public Sketch(String locatorURI, BitSet bitSet) {
        super(locatorURI);
        this.bitSet = bitSet;
    }

    public Sketch(BitSet bitSet) {
        this.bitSet = bitSet;
    }

    public Sketch(Sketch sketchToCopy) {
        this(sketchToCopy.getLocatorURI(), (BitSet) sketchToCopy.bitSet.clone());
    }

    // predpokladam stejnou delku prefixu pro vsechna pouziti
    public Sketch(Sketch prefix, Sketch suffix) {
        super(prefix.getLocatorURI());
        long[] prefixLongs = prefix.bitSet.toLongArray();
        long[] suffixLongs = suffix.bitSet.toLongArray();
        long[] concat = new long[prefixLongs.length + suffixLongs.length];
        System.arraycopy(prefixLongs, 0, concat, 0, prefixLongs.length);
        System.arraycopy(suffixLongs, 0, concat, prefixLongs.length, suffixLongs.length);
        bitSet = BitSet.valueOf(concat);
    }

    /**
     * Constructor. Consider to use constructor with specified LocatorURI.
     * (Locator is final atribut of class.)
     *
     * @param size Number of allocated bits when Sketch is created. Bits are
     * automatically added in case of insufficient length. All bits are set to
     * 0.
     */
    public Sketch(int size) {
        bitSet = new BitSet(size);
    }

    /**
     * Constructor used for reading the stored Sketch.
     *
     * @param br BufferedReader used to read from a file.
     * @throws IOException
     */
    public Sketch(BufferedReader br) throws IOException {
        String[] longS = readObjectComments(br).split(",");
        long[] longs = new long[longS.length];
        for (int i = 0; i < longS.length; i++) {
            longs[i] = Long.parseLong(longS[i]);
        }
        this.bitSet = BitSet.valueOf(longs);
    }

    /**
     * Constructor
     *
     * @param locatorURI locator of the sketch (usually the same as the
     * constructor of the original data object).
     * @param size Number of allocated bits when Sketch is created. Bits are
     * automatically added in case of insufficient length. All bits are set to
     * 0.
     */
    public Sketch(String locatorURI, int size) {
        super(locatorURI);
        bitSet = new BitSet(size);
    }

    /**
     * Constructor
     *
     * @param locatorURI locator of the sketch (usually the same as the
     * constructor of the original data object).
     */
    public Sketch(String locatorURI) {
        super(locatorURI);
    }

    /**
     * Sets the bit value in specified position to be zero. (Not neccesary when
     * a new Sketch is created.)
     *
     * @param position bit position
     */
    public void setBitValueNegative(int position) {
        bitSet.set(position, false);
    }

    /**
     * Sets the bit value in specified position to be one.
     *
     * @param position bit position
     */
    public void setBitValuePositive(int position) {
        bitSet.set(position);
    }

    public boolean getBitValue(int position) {
        return bitSet.get(position);
    }

    public boolean isInIntersectOf(int... intersectingPositions) {
        for (int pos : intersectingPositions) {
            if (!bitSet.get(pos)) {
                return false;
            }
        }
        return true;
    }

    @Override
    public int getSize() {
        return bitSet.size() / 8;
    }

    @Override
    public int hashCode() {
        int hash = 7;
        hash = 59 * hash + Objects.hashCode(this.bitSet);
        return hash;
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Sketch other = (Sketch) obj;
        return Objects.equals(this.bitSet, other.bitSet);
    }

    /**
     * Without locator
     *
     * @param otherBitSet
     * @return
     */
    public BitSet getNewXoredBitSet(BitSet otherBitSet) {
        BitSet thisCloned = (BitSet) bitSet.clone();
        thisCloned.xor(otherBitSet);
        return thisCloned;
    }

    @Override
    protected float getDistanceImpl(LocalAbstractObject obj, float distThreshold) {
        // Hamming distance
        Sketch cast = (Sketch) obj;
        BitSet thisCloned = (BitSet) bitSet.clone();
        thisCloned.xor(cast.bitSet);
//        for (int i = 0; i < thisCloned.length(); i++) {
//            if (thisCloned.get(i)) {
//                System.out.print("1;");
//            } else {
//                System.out.print("0;");
//            }
//        }
//        System.out.println();
        return thisCloned.cardinality();
    }

    @Override
    public boolean dataEquals(Object obj) {
        Sketch skObj = (Sketch) obj;
        return bitSet.equals(skObj.bitSet);
    }

    @Override
    public int dataHashCode() {
        return bitSet.hashCode();
    }

    @Override
    protected void writeData(OutputStream stream) throws IOException {
        long[] data = bitSet.toLongArray();
        if (data.length == 0) {
            stream.write('0');
        } else {
            for (int i = 0; i < data.length - 1; i++) {
                stream.write(Long.toString(data[i]).getBytes());
                stream.write(',');
            }
            stream.write(Long.toString(data[data.length - 1]).getBytes());
        }
        stream.write('\n');
    }

    @Override
    public int compareTo(Sketch o) {
        return (int) getDistance(o);
    }

    /**
     * Returns values used during the evaluation ofthe Pearson correlation
     * coeficient.
     *
     * @param sketchLength The length of sketch. Set this value manually, since
     * it does not have to correspond to any method called on Sketch.
     * @return array with: number of ones in sketch, number of ones multiplied
     * with sketch length and second power of number of ones.
     */
    public float[] getCardSizeCardCard2(int sketchLength) {
        float[] ret = new float[3];
        ret[0] = getCardinality();
//        if (ret[0] < sketchLength - ret[0]) {
//            ret[0] = sketchLength - ret[0];
//            bitSet.flip(0, sketchLength);
//        }
        ret[1] = sketchLength * ret[0];
        ret[2] = ret[0] * ret[0];
        return ret;
    }

    /**
     * Evaluate the Pearson correlation coefficient of two sketches.
     * Precomputedd values must be provided (see description of parameters).
     *
     * @param otherSketch Sketch (i.e. bit-vector) to be compared with this one.
     * @param thisCards result of method getCardSizeCardCard2 called on this
     * sketch
     * @param otherCards result of method getCardSizeCardCard2 called on
     * otherSketch
     * @param sketchLength The length of sketch. Set this value manually, since
     * it does not have to correspond to any method called on Sketch.
     * @return Pearson correlation value on tis Sketch and other sketch
     */
    public float getPearsonCorrelation(Sketch otherSketch, float[] thisCards, float[] otherCards, long sketchLength) {
//        boolean oriented = bitSet.cardinality() >= sketchLength - bitSet.cardinality();
//        oriented = oriented && otherSketch.bitSet.cardinality() >= sketchLength - otherSketch.bitSet.cardinality();
//        if (!oriented) {
//            throw new RuntimeException("Not oriented");
//        }
        double aCard = thisCards[0];
        double aCardSize = thisCards[1];
        double aCard2 = thisCards[2];
        double bCard = otherCards[0];
        double bCardSize = otherCards[1];
        double bCard2 = otherCards[2];

        BitSet bitSet1 = (BitSet) bitSet.clone();

        bitSet1.and(otherSketch.bitSet);
        int a = bitSet1.cardinality();

        double numerator = (sketchLength * a) - (aCard * bCard);
        double denominator = Math.sqrt((aCardSize - aCard2) * (bCardSize - bCard2));
        double ret = numerator / denominator;
        if (ret > 1) {
            Logger.getLogger(Sketch.class.getName()).log(Level.SEVERE,
                    "Wrong correlation! {0};{1};{2}\n{3};{4};{5}\nnumerator: {6}; denominator: {7}\n\n",
                    new Object[]{aCard, aCardSize, aCard2, bCard, bCardSize, bCard2, numerator, denominator});
        }
        return (float) ret;
    }

    public int getCardinality() {
        return bitSet.cardinality();
    }

    @Deprecated
    public boolean[] getAsBoolArray() {
        boolean[] ret = new boolean[bitSet.size()];
        for (int i = 0; i < bitSet.size(); i++) {
            ret[i] = bitSet.get(i);
        }
        return ret;
    }

    @Deprecated
    public int[] getAsIntArray() {
        int[] ret = new int[bitSet.size()];
        for (int i = 0; i < bitSet.size(); i++) {
            ret[i] = bitSet.get(i) ? 1 : 0;
        }
        return ret;
    }

    public long[] getAsLongArray() {
        return bitSet.toLongArray();
    }

    public byte[] getAsByteArray() {
        return bitSet.toByteArray();
    }

    @Deprecated
    public String getAsReadableString() {
        String ret = "";
        boolean[] asBoolArray = getAsBoolArray();
        for (int i = 0; i < asBoolArray.length; i++) {
            if (asBoolArray[i]) {
                ret += "1";
            } else {
                ret += "0";
            }
        }
        return ret;
    }

    @Deprecated
    public static boolean[] getAsBoolArray(BitSet bitSet) {
        boolean[] ret = new boolean[bitSet.size()];
        for (int i = 0; i < bitSet.size(); i++) {
            ret[i] = bitSet.get(i);
        }
        return ret;
    }

    public void performXorOnBitset(BitSet bitSet) {
        bitSet.xor(this.bitSet);
    }

    @Deprecated
    public boolean[] getXoredSketchAsBools(Sketch obj) {
        Sketch cast = (Sketch) obj;
        BitSet thisCloned = (BitSet) bitSet.clone();
        thisCloned.xor(cast.bitSet);
        return getAsBoolArray(thisCloned);
    }

    public int[] getCardinalitiesInBlocks(int numberOfBlocks, int sketchLength) {
        int[] ret = new int[numberOfBlocks];
        int blockSize = sketchLength / numberOfBlocks;
        for (int i = 0; i < numberOfBlocks; i++) {
            int start = i * blockSize;
            int end = start + blockSize;
            if (end + blockSize > sketchLength) {
                end = sketchLength;
            }

//            for (int c = bitSet.nextSetBit(start); c >= 0 && c < end; c = bitSet.nextSetBit(c + 1)) {
//                ret[i]++;
//            }
            ret[i] = bitSet.get(start, end).cardinality();
        }
        return ret;
    }

    public Sketch[] getTwoSubSketches(int lengthOfFirst) {
        Sketch prefix = new Sketch(getLocatorURI(), bitSet.get(0, lengthOfFirst));
        Sketch suffix = new Sketch(bitSet.get(lengthOfFirst + 1, Integer.MAX_VALUE));
        return new Sketch[]{prefix, suffix};
    }

    public void switchBit(int idx) {
        bitSet.flip(idx);
    }

}
