package sf.objects.sketchcreator;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import messif.objects.LocalAbstractObject;
import messif.objects.util.AbstractObjectList;
import messif.objects.util.StreamGenericAbstractObjectIterator;
import sf.objects.impl.Sketch;
import sf.objects.sketchcreator.impl.GenHyperplane;
import sf.pivotAnalyzer.uncorrelatedSelector.InvertedSketchesCorrelationCounter;
import sf.pivotAnalyzer.uncorrelatedSelectorImpl.HeuristicCorrelationSubsetSearch;

/**
 *
 * @author xmic
 */
public abstract class ObjectToSketchTransformator implements Serializable {

    public static final long serialVersionUID = 736012244806410000l;
    public static final Integer PARALELISM = 8;
    public static final Float BALANCE_TOLERATION = 0.05f;

    protected LocalAbstractObject[] pivots;

    private int sampleObjectNumber = 0;

    public ObjectToSketchTransformator(LocalAbstractObject[] pivots) {
        this.pivots = pivots;
    }

    public ObjectToSketchTransformator(String pivotFile, String objectClass, int pivotCount) {
        this.pivots = readPivots(pivotFile, objectClass, pivotCount);
    }

    private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
    }

    protected final LocalAbstractObject[] readPivots(String pivotFile, String objectClass, int pivotCount) {
        try {
            Iterator<LocalAbstractObject> it = new StreamGenericAbstractObjectIterator(Class.forName(objectClass), pivotFile);
            return new AbstractObjectList<>(it, pivotCount).toArray(new LocalAbstractObject[0]);
        } catch (IllegalArgumentException | IOException | ClassNotFoundException ex) {
            Logger.getLogger(GenHyperplane.class.getName()).log(Level.SEVERE, null, ex);
        }
        return null;
    }

    public void removeCorrelatedBits(int finalBitsNumber, String sampleSetFile, String objectClass) {
        this.removeCorrelatedBits(finalBitsNumber, sampleSetFile, objectClass, 0);
    }

    public void removeCorrelatedBits(int finalBitsNumber, String sampleSetFile, String objectClass, float correlationToAchieve) {
        try {
            Iterator<LocalAbstractObject> sampleSet = new StreamGenericAbstractObjectIterator(Class.forName(objectClass), sampleSetFile);
            int[] uncorrelatedBits = getUncorrelatedBits(finalBitsNumber, sampleSet, correlationToAchieve);
            removeCorrelatedBits(uncorrelatedBits);
        } catch (ClassNotFoundException | IllegalArgumentException | IOException ex) {
            Logger.getLogger(GenHyperplane.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    public List<Sketch> createInvertedSketches(Iterator<LocalAbstractObject> sampleSet) {
        try {
            List<Sketch> ret = new ArrayList<>();
            int invertedSketchesCount = getInvertedSketchesCount();
            for (int i = 0; i < invertedSketchesCount; i++) {
                ret.add(new Sketch(0));
            }
            int threadId = 0;
            InvertedSketchesCreatorThread[] threads = new InvertedSketchesCreatorThread[PARALELISM];
            sampleObjectNumber = 0;
            int firstObject = 0;
            while (sampleSet.hasNext()) {
                AbstractObjectList<LocalAbstractObject> batch = new AbstractObjectList(sampleSet, 500);
                int threadIdx = threadId % PARALELISM;
                threads[threadIdx] = new InvertedSketchesCreatorThread(batch, threadId, ret, this, firstObject);
                threads[threadIdx].start();
                firstObject += batch.size();
                if (threadIdx == PARALELISM - 1) {
                    for (int i = 0; i < PARALELISM; i++) {
                        threads[i].join();
                        sampleObjectNumber += threads[i].getAndResetSampleObjectNumber();
                    }
                }
                threadId++;
            }
            for (int i = 0; i < PARALELISM; i++) {
                if (threads[i] != null) {
                    threads[i].join();
                    sampleObjectNumber += threads[i].getAndResetSampleObjectNumber();
                }
            }
            Logger.getLogger(GenHyperplane.class.getName()).log(Level.INFO, "Inverted sketches Processed (sample objects count {0})", new Object[]{sampleObjectNumber});
            return ret;
        } catch (InterruptedException ex) {
            Logger.getLogger(ObjectToSketchTransformator.class.getName()).log(Level.SEVERE, null, ex);
        }
        return null;
    }

    private int[] getUncorrelatedBits(int tuplesLength, int finalBitsNumber, Iterator<LocalAbstractObject> sampleSet, float correlationToAchieve) {
        if (pivots != null && finalBitsNumber >= pivots.length) {
            return null;
        }
        List<Sketch> invertedSketches = createInvertedSketches(sampleSet);
        return getUncorrelatedBits(tuplesLength, finalBitsNumber, invertedSketches, correlationToAchieve);
    }

    private int[] getUncorrelatedBits(int tuplesLength, int finalBitsNumber, List<Sketch> invertedSketches, float correlationToAchieve) {
        InvertedSketchesCorrelationCounter correlationCounter = new HeuristicCorrelationSubsetSearch(invertedSketches, sampleObjectNumber, correlationToAchieve);
        return correlationCounter.getIndexesOfLowCorrelatedPivots(tuplesLength, finalBitsNumber);
    }

    private int[] getUncorrelatedBits(int finalBitsNumber, List<Sketch> invertedSketches, float correlationToAchieve) {
        InvertedSketchesCorrelationCounter correlationCounter = new HeuristicCorrelationSubsetSearch(invertedSketches, sampleObjectNumber, correlationToAchieve);
        return correlationCounter.getIndexesOfLowCorrelatedPivots(1, finalBitsNumber);
    }

    protected int[] getUncorrelatedBits(int finalBitsNumber, Iterator<LocalAbstractObject> sampleSet, float correlationToAchieve) {
        return getUncorrelatedBits(1, finalBitsNumber, sampleSet, correlationToAchieve);
    }

    protected int[] getUncorrelatedBitPairs(int finalBitsNumber, Iterator<LocalAbstractObject> sampleSet) {
        return getUncorrelatedBits(2, finalBitsNumber, sampleSet, 0);
    }

    /**
     * Methods used for the constructor InvertedSketchesCorrelationCounter
     *
     * @return
     */
    public int getSampleObjectNumber() {
        return sampleObjectNumber;
    }

    /**
     * Create sketch fot the object
     *
     * @param objectToTransform
     * @return Sketch of the provided object
     */
    public abstract Sketch transform(LocalAbstractObject objectToTransform);

    /**
     * Create inverted sketches (i.e. bit-strings describing the columns of the
     * matrix made by sketches - rows are the sketches.) Result is stored in
     * parameter invertedSketches.
     *
     * @param sampleObj Object from which is created the sketch an the matrix is
     * updated. I.e. in all columns is set the proper value according to this
     * sketch.
     * @param invertedSketches List of inverted sketches (matrix representation)
     * - the result is stored here. At the really beggining it should contain
     * the proper number of sketches with all zeros.
     * @param objectNumber number of matrix rows
     */
    public abstract void transformToInvertedFile(LocalAbstractObject sampleObj, List<Sketch> invertedSketches, int objectNumber);

//    public abstract void removeCorrelatedBits(int finalSketchLength, Iterator<LocalAbstractObject> sampleSet);
    public abstract void removeCorrelatedBits(int[] uncorrelatedBits);

    public abstract int getInvertedSketchesCount();

    public int getSketchLengths() {
        return getInvertedSketchesCount();
    }

    public List<Sketch> transformAllObjectsToSketchesComplete(Iterator<LocalAbstractObject> dataObjects, List<LocalAbstractObject> sampleObjects, float balance, int sketchLength, int balancedBitsCount, String destPath) {
        return this.transformAllObjectsToSketchesComplete(dataObjects, sampleObjects, balance, sketchLength, balancedBitsCount, 0, destPath);
    }

    public List<Sketch> transformAllObjectsToSketchesComplete(Iterator<LocalAbstractObject> dataObjects, List<LocalAbstractObject> sampleObjects, float balance, int sketchLength, int balancedBitsCount, float correlationToAchieve, String destPath) {
        if (balance < 0 || balance > 1) {
            throw new IllegalArgumentException("Set balanced from range (0, 1).");
        }
        List<Sketch> invertedSketches = createInvertedSketches(sampleObjects.iterator());
        List<Integer> indexesOfBalancedInvertedSketches = filterBalancedSketches(invertedSketches, balance, sampleObjects.size(), balancedBitsCount);
        invertedSketches = filterList(invertedSketches, indexesOfBalancedInvertedSketches, balancedBitsCount);
        int[] uncorrelatedBits = getUncorrelatedBits(sketchLength, invertedSketches, correlationToAchieve);
        uncorrelatedBits = transformIndexesOfUncorrelated(uncorrelatedBits, indexesOfBalancedInvertedSketches);
        removeCorrelatedBits(uncorrelatedBits);
        sampleObjects.clear();
        System.gc();
        List<Sketch> ret = new ArrayList<>();
        boolean print = false;
        if (destPath != null) {
            print = true;
            try {
                System.setOut(new PrintStream(destPath));
            } catch (FileNotFoundException ex) {
                Logger.getLogger(ObjectToSketchTransformator.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        for (int i = 0; dataObjects.hasNext(); i++) {
            LocalAbstractObject next = dataObjects.next();
            Sketch sk = transform(next);
            if (print) {
                try {
                    sk.write(System.out);
                } catch (IOException ex) {
                    Logger.getLogger(ObjectToSketchTransformator.class.getName()).log(Level.SEVERE, null, ex);
                }
            } else {
                ret.add(sk);
            }
        }
        if (print) {
            System.out.flush();
        }
        return ret;
    }

    private List<Integer> filterBalancedSketches(List<Sketch> allInvertedSketches, float balance, int sampleObjectCounts, int balancedBitsCount) {
        List<Integer> indexes = new ArrayList<>();
        if (balance == 0 || balance == 1) {
            for (int i = 0; i < allInvertedSketches.size(); i++) {
                indexes.add(i);

            }
            Logger.getLogger(ObjectToSketchTransformator.class
                    .getName()).log(Level.INFO, "All {0} bits are used (ignoring their balance).", new Object[]{indexes.size()
                    }
                    );
        } else {
            int[] cardinalities = new int[allInvertedSketches.size()];
            for (int i = 0; i < allInvertedSketches.size(); i++) {
                cardinalities[i] = allInvertedSketches.get(i).getCardinality();
            }
            int min = (int) ((balance - BALANCE_TOLERATION) * sampleObjectCounts);
            int max = (int) ((balance + BALANCE_TOLERATION) * sampleObjectCounts);
            for (int i = 0; i < cardinalities.length; i++) {
                if (cardinalities[i] <= max && cardinalities[i] >= min) {
                    indexes.add(i);
                } else {
                    int reversedCard = sampleObjectCounts - cardinalities[i];
                    if (reversedCard <= max && reversedCard >= min) {
                        if (this instanceof GenHyperplane) {
                            GenHyperplane tr = (GenHyperplane) this;
                            tr.reorderPivotsToSwitchBits(i);
                            indexes.add(i);
                        } else {
                            throw new RuntimeException("Need to switch pivots to set bit but stange ObjectToSketchTransformator provided.");

                        }
                    }
                }
            }
            if (indexes.size() < balancedBitsCount) {
                Logger.getLogger(ObjectToSketchTransformator.class
                        .getName()).log(Level.WARNING, "Only {0} bits balanced up to {1} were found.", new Object[]{indexes.size(), balance});

            } else {
                Logger.getLogger(ObjectToSketchTransformator.class
                        .getName()).log(Level.INFO, "{0} bits balanced up to {1} were found. {2} will be selected.", new Object[]{indexes.size(), balance, balancedBitsCount});
            }
        }
//permutate?
        Collections.shuffle(indexes);
//permutate?

        return indexes;
    }

    private List filterList(List allInvertedSketches, List<Integer> indexes, int count) {
        List ret = new ArrayList();
        for (int i = 0; i < Math.min(indexes.size(), count); i++) {
            ret.add(allInvertedSketches.get(indexes.get(i)));
        }
        return ret;
    }

    private int[] transformIndexesOfUncorrelated(int[] uncorrelatedBits, List<Integer> indexesOfBalancedInvertedSketches) {
        int[] ret = new int[uncorrelatedBits.length];
        for (int i = 0; i < uncorrelatedBits.length; i++) {
            int idx = uncorrelatedBits[i];
            ret[i] = indexesOfBalancedInvertedSketches.get(idx);
        }
        return ret;
    }

    public static List<Sketch> tranformSketchesToInvertedSketches(List<Sketch> sketches) {
        List<Sketch> ret = new ArrayList<>();
        long numberOfOnesCheckSum = 0;
        for (Sketch sk : sketches) {
            numberOfOnesCheckSum += sk.getCardinality();
        }
        long setOnes = 0;
        int bitIdx = 0;
        while (setOnes != numberOfOnesCheckSum) {
            for (int i = 0; i < sketches.size(); i++) {
                if (sketches.get(i).getBitValue(bitIdx)) {
                    while (ret.size() <= bitIdx) {
                        ret.add(new Sketch(0));
                    }
                    ret.get(bitIdx).setBitValuePositive(i);
                    setOnes++;
                }
            }
            bitIdx++;
        }
        return ret;
    }

}
